"use strict";

(function() {

Components.utils.import("resource://gre/modules/Services.jsm");

const Cc = Components.classes, Ci = Components.interfaces;

const kDateFormatNone = 0;              // do not include the date  in the format string
const kDateFormatLong = 1;              // provides the long date format for the given locale
const kDateFormatShort = 2;             // provides the short date format for the given locale
const kDateFormatYearMonth = 3;         // formats using only the year and month
const kDateFormatWeekday = 4;           // week day (e.g. Mon, Tue)

if (!MailredirectExtension)
  var MailredirectExtension = {};

var dumper = new MailredirectDebug.Dump();

MailredirectExtension.resentFromColumnHandler =
{
  getCellText: function(row, col) {
    // get the message's header so that we can extract the resent-from fields
    var hdr = gDBView.getMsgHdrAt(row);
    var retval = hdr.getStringProperty("resent-from");
    return retval;
  },
  getSortStringForRow: function(hdr)
  {
    return hdr.getStringProperty("resent-from");
  },
  isString: function() { return true; },
  getCellProperties: function(row, col, props) {},
  getRowProperties: function(row, props) {},
  getImageSrc: function(row, col) { return null; },
  getSortLongForRow: function(hdr) { return 0; }
};

MailredirectExtension.resentDateColumnHandler =
{
  getCellText: function(row, col) {
    // get the message's header so that we can extract the resent-date fields
    var hdr = gDBView.getMsgHdrAt(row);
    var date = Date.parse(hdr.getStringProperty("resent-date"));
    dumper.dump("date: " + date);
    var retval = hdr.getStringProperty("resent-date");
    return retval;
  },
  getSortStringForRow: function(hdr)
  {
    return Date.parse(hdr.getStringProperty("resent-date"));
  },
  isString: function() { return true; },
  getCellProperties: function(row, col, props) {},
  getRowProperties: function(row, props) {},
  getImageSrc: function(row, col) { return null; },
  getSortLongForRow: function(hdr) { return 0; }
};

MailredirectExtension.CreateDbObserver =
{
  // Components.interfaces.nsIObserver
  observe: function(aMsgFolder, aTopic, aData)
  {
    gDBView.addColumnHandler("resentFromCol", MailredirectExtension.resentFromColumnHandler);
    gDBView.addColumnHandler("resentDateCol", MailredirectExtension.resentDateColumnHandler);
  }
};

MailredirectExtension.AddDBObserver = function()
{
  var observerService = Cc["@mozilla.org/observer-service;1"].
                        getService(Ci.nsIObserverService);
  observerService.addObserver(MailredirectExtension.CreateDbObserver, "MsgCreateDBView", false);
};

MailredirectExtension.RemoveDBObserver = function()
{
  var observerService = Cc["@mozilla.org/observer-service;1"].
                        getService(Ci.nsIObserverService);
  observerService.removeObserver(MailredirectExtension.CreateDbObserver, "MsgCreateDBView");
};

MailredirectExtension.InitializeColumnOverlay = function()
{
  function getDateFormatPref(prefBranch, prefLocalName, defaultFormat)
  {
    var retval = defaultFormat;
    try {
      var result = prefBranch.getIntPref(prefLocalName);
      if (result >= kDateFormatShort && result <= kDateFormatWeekday) retval = result;
    } catch(ex) { }
    return retval;
  };

  var customDBHeaders = Services.prefs.getCharPref("mailnews.customDBHeaders");
  var newHeaders = customDBHeaders.split(" ");
  if (newHeaders.indexOf("resent-from") === -1) newHeaders.push("resent-from");
  if (newHeaders.indexOf("resent-date") === -1) newHeaders.push("resent-date");
  newHeaders = newHeaders.join(" ");
  if (newHeaders !== customDBHeaders) Services.prefs.setCharPref("mailnews.customDBHeaders", newHeaders);

  var dateFormatPrefs = Services.prefs.getBranch("mail.ui.display.dateformat.");
  var m_dateFormatDefault = getDateFormatPref(dateFormatPrefs, "default", kDateFormatShort);
  var m_dateFormatThisWeek = getDateFormatPref(dateFormatPrefs, "thisweek", kDateFormatShort);
  var m_dateFormatToday = getDateFormatPref(dateFormatPrefs, "today", kDateFormatNone);
};

window.addEventListener("load", MailredirectExtension.AddDBObserver);
window.addEventListener("unload", MailredirectExtension.RemoveDBObserver);

MailredirectExtension.InitializeColumnOverlay();

})();
